"""
Tests for `typed_settings.attrs.hooks`.
"""
import functools
import json
from datetime import datetime
from enum import Enum
from typing import (
    Any,
    Dict,
    FrozenSet,
    List,
    Mapping,
    MutableMapping,
    MutableSequence,
    MutableSet,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
)

import attr
import pytest

from typed_settings.attrs.hooks import auto_convert, auto_serialize
from typed_settings.attrs.validators import le


auto_converter = functools.partial(attr.frozen, field_transformer=auto_convert)


class TestTransformHook:
    """
    Tests for `attrs(tranform_value_serializer=func)`
    """

    def test_hook_applied(self):
        """
        The transform hook is applied to all attributes.  Types can be missing,
        explicitly set, or annotated.
        """
        results = []

        def hook(cls, attribs):
            results[:] = [(a.name, a.type) for a in attribs]
            return attribs

        @attr.s(field_transformer=hook)
        class C:
            x = attr.ib()
            y = attr.ib(type=int)
            z: float = attr.ib()

        assert results == [("x", None), ("y", int), ("z", float)]

    def test_hook_applied_auto_attrib(self):
        """
        The transform hook is applied to all attributes and type annotations
        are detected.
        """
        results = []

        def hook(cls, attribs):
            results[:] = [(a.name, a.type) for a in attribs]
            return attribs

        @attr.s(auto_attribs=True, field_transformer=hook)
        class C:
            x: int
            y: str = attr.ib()

        assert results == [("x", int), ("y", str)]

    def test_hook_applied_modify_attrib(self):
        """
        The transform hook can modify attributes.
        """

        def hook(cls, attribs):
            return [a.evolve(converter=a.type) for a in attribs]

        @attr.s(auto_attribs=True, field_transformer=hook)
        class C:
            x: int = attr.ib(converter=int)
            y: float

        c = C(x="3", y="3.14")
        assert c == C(x=3, y=3.14)

    def test_hook_remove_field(self):
        """
        It is possible to remove fields via the hook.
        """

        def hook(cls, attribs):
            return [a for a in attribs if a.type is not int]

        @attr.s(auto_attribs=True, field_transformer=hook)
        class C:
            x: int
            y: float

        assert attr.asdict(C(2.7)) == {"y": 2.7}

    def test_hook_add_field(self):
        """
        It is possible to add fields via the hook.
        """

        def hook(cls, attribs):
            a1 = attribs[0]
            a2 = a1.evolve(name="new")
            return [a1, a2]

        @attr.s(auto_attribs=True, field_transformer=hook)
        class C:
            x: int

        assert attr.asdict(C(1, 2)) == {"x": 1, "new": 2}

    def test_hook_with_inheritance(self):
        """
        The hook receives all fields from base classes.
        """

        def hook(cls, attribs):
            assert [a.name for a in attribs] == ["x", "y"]
            # Remove Base' "x"
            return attribs[1:]

        @attr.s(auto_attribs=True)
        class Base:
            x: int

        @attr.s(auto_attribs=True, field_transformer=hook)
        class Sub(Base):
            y: int

        assert attr.asdict(Sub(2)) == {"y": 2}


class TestAsDictHook:
    def test_asdict(self):
        """
        asdict() calls the hooks in attrs classes and in other datastructures
        like lists or dicts.
        """

        def hook(inst, a, v):
            if isinstance(v, datetime):
                return v.isoformat()
            return v

        @attr.dataclass
        class Child:
            x: datetime
            y: List[datetime]

        @attr.dataclass
        class Parent:
            a: Child
            b: List[Child]
            c: Dict[str, Child]
            d: Dict[str, datetime]

        inst = Parent(
            a=Child(1, [datetime(2020, 7, 1)]),
            b=[Child(2, [datetime(2020, 7, 2)])],
            c={"spam": Child(3, [datetime(2020, 7, 3)])},
            d={"eggs": datetime(2020, 7, 4)},
        )

        result = attr.asdict(inst, value_serializer=hook)
        assert result == {
            "a": {"x": 1, "y": ["2020-07-01T00:00:00"]},
            "b": [{"x": 2, "y": ["2020-07-02T00:00:00"]}],
            "c": {"spam": {"x": 3, "y": ["2020-07-03T00:00:00"]}},
            "d": {"eggs": "2020-07-04T00:00:00"},
        }

    def test_asdict_calls(self):
        """
        The correct instances and attribute names are passed to the hook.
        """
        calls = []

        def hook(inst, a, v):
            calls.append((inst, a.name if a else a, v))
            return v

        @attr.dataclass
        class Child:
            x: int

        @attr.dataclass
        class Parent:
            a: Child
            b: List[Child]
            c: Dict[str, Child]

        inst = Parent(a=Child(1), b=[Child(2)], c={"spam": Child(3)})

        attr.asdict(inst, value_serializer=hook)
        assert calls == [
            (inst, "a", inst.a),
            (inst.a, "x", inst.a.x),
            (inst, "b", inst.b),
            (inst.b[0], "x", inst.b[0].x),
            (inst, "c", inst.c),
            (None, None, "spam"),
            (inst.c["spam"], "x", inst.c["spam"].x),
        ]


class LeEnum(Enum):
    spam = "Le Spam"
    eggs = "Le Eggs"


@auto_converter
class Child:
    x: int = attr.ib()
    y: int = attr.ib(converter=int)


@auto_converter(kw_only=True)
class Parent:
    child: Child
    a: float
    c: LeEnum
    d: datetime
    e: "List[Child]"
    f: Set[datetime]
    b: float = attr.field(default=3.14, validator=le(2))


class TestAutoConvertHook:
    """Tests for the bundled auto-convert hook."""

    DATA = {
        "a": "1",
        "b": "2",
        "c": "spam",
        "d": "2020-05-04T13:37:00",
        "e": [{"x": "23", "y": "42"}],
        "f": ["2020-05-04T13:37:00", "2020-05-04T13:37:00"],
        "child": {"x": 23, "y": "42"},  # Also tests int->str conversion
    }

    @pytest.fixture(scope="class")
    def parent(self):
        return Parent(**self.DATA)

    def test_auto_convert_hook(self, parent):
        """
        The auto_convert hook converts attrs classes, datetimes, enums and
        basic type as well as basic containers.

        This is a basic test to assert that this generally works.

        The tests "test_static_types()" and "test_generic_types()" test the
        full list of supported types.
        """
        assert parent == Parent(
            a=1.0,
            b=2.0,
            c=LeEnum.spam,
            d=datetime(2020, 5, 4, 13, 37),
            e=[Child(23, 42)],
            f={datetime(2020, 5, 4, 13, 37)},
            child=Child(23, 42),
        )

    def test_serialize_from_parent(self, parent):
        """
        The serialize_hook is able to serialize the same types as the
        auto_convert hook.

        The set with duplicate entries of attrib "f" is not the same as in the
        original dict!
        """
        d = attr.asdict(parent, value_serializer=auto_serialize)
        assert d == {
            "a": 1.0,
            "b": 2.0,
            "c": "spam",
            "d": "2020-05-04T13:37:00",
            "e": [{"x": 23, "y": 42}],
            "f": ["2020-05-04T13:37:00"],
            "child": {"x": 23, "y": 42},
        }

    def test_json_roundtrip(self, parent):
        """
        The roundtrip "inst -> JSON -> inst" results in the same instance.
        """
        d = attr.asdict(parent, value_serializer=auto_serialize)
        assert Parent(**json.loads(json.dumps(d))) == parent

    @pytest.mark.parametrize(
        "typ, value, expected",
        [
            (bool, True, True),
            (bool, "True", True),
            (bool, 1, True),
            (bool, False, False),
            (bool, "False", False),
            (bool, 0, False),
            (int, 23, 23),
            (int, "42", 42),
            (float, 3.14, 3.14),
            (float, ".815", 0.815),
            (str, "spam", "spam"),
            (datetime, "2020-05-04T13:37:00", datetime(2020, 5, 4, 13, 37)),
            (LeEnum, "eggs", LeEnum.eggs),
            (Child, {"x": "2", "y": "3"}, Child(2, 3)),
            (Any, 2, 2),
            (Any, "2", "2"),
            (Any, None, None),
        ],
    )
    def test_static_types(self, typ, value, expected):
        """
        All oficially supported types can be converted by attrs.

        Please create an issue if something is missing here.
        """

        @auto_converter
        class C:
            opt: typ

        c = C(value)
        assert c.opt == expected
        if typ is not Any:
            assert type(c.opt) is typ

    @pytest.mark.parametrize(
        "typ, value, expected_val, expected_type",
        [
            (List[int], [1, 2], [1, 2], list),
            (List[Child], [{"x": 1, "y": 2}], [Child(1, 2)], list),
            (
                List[Any],
                [True, None, 23, 3.14, "spam"],
                [True, None, 23, 3.14, "spam"],
                list,
            ),
            (Sequence[int], [1, 2], [1, 2], list),
            (MutableSequence[int], [1, 2], [1, 2], list),
            (Set[int], [1, 2], {1, 2}, set),
            (MutableSet[int], [1, 2], {1, 2}, set),
            (FrozenSet[int], [1, 2], frozenset({1, 2}), frozenset),
            (Tuple[str, ...], [1, "2", 3], ("1", "2", "3"), tuple),
            (Tuple[int, bool, str], [0, "no", 0], (0, False, "0"), tuple),
            (Dict[str, int], {"a": 1, "b": 3.1}, {"a": 1, "b": 3}, dict),
            (
                Dict[str, Child],
                {"a": {"x": "1", "y": "2"}},
                {"a": Child(1, 2)},
                dict,
            ),
            (
                Dict[Tuple[int, int], List[Dict[int, Child]]],
                {
                    ("1", "2"): [
                        {"3": {"x": "4", "y": "5"}},
                        {"6": {"x": "7", "y": "8"}},
                    ],
                },
                {(1, 2): [{3: Child(4, 5)}, {6: Child(7, 8)}]},
                dict,
            ),
            (Mapping[str, int], {"a": 1, "b": 3.1}, {"a": 1, "b": 3}, dict),
            (
                MutableMapping[str, int],
                {"a": 1, "b": 3.1},
                {"a": 1, "b": 3},
                dict,
            ),
            (Optional[str], 1, "1", str),
            (Optional[Child], None, None, type(None)),
            (Optional[Child], {"x": "1", "y": "2"}, Child(1, 2), Child),
            (Optional[LeEnum], "spam", LeEnum.spam, LeEnum),
            (Union[None, Child], None, None, type(None)),
            (Union[None, Child], {"x": "1", "y": "2"}, Child(1, 2), Child),
            (Union[None, List[str]], None, None, type(None)),
            (Union[None, List[str]], [1, 2], ["1", "2"], list),
        ],
    )
    def test_generic_types(self, typ, value, expected_val, expected_type):
        """
        All oficially supported types can be converted by attrs.

        Please create an issue if something is missing here.
        """

        @auto_converter
        class C:
            opt: typ

        c = C(value)
        assert c.opt == expected_val
        assert type(c.opt) is expected_type



try:
    from importlib.metadata import version as get_version
except ImportError:
    # py37 - Just create a mock for testing:
    def get_version(_):
        return '0.0.0'


project = "Typed Settings"
author = "Stefan Scherfke"
copyright = "2020, Stefan Scherfke"
release = get_version("typed-settings")
version = ".".join(release.split(".")[0:2])


extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx_autodoc_typehints",
]
templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


html_theme = "furo"
html_theme_options = {
    # "logo_only": True,
    # "sidebar_hide_name": True,
    "light_css_variables": {
        "color-brand-primary": "#266DB4",
        "color-brand-content": "#266DB4",
    },
    "dark_css_variables": {
        "color-brand-primary": "#3186DC",
        "color-brand-content": "#3186DC",
    },
}
html_static_path = ["_static"]
html_logo = "_static/typed-settings-spacing.svg"
html_title = "Typed Settings"


# Autodoc
autodoc_member_order = "bysource"

# Intersphinx
intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "attrs": ("https://www.attrs.org/en/stable/", None),
    "cattrs": ("https://cattrs.readthedocs.io/en/latest/", None),
    "click": ("https://click.palletsprojects.com/en/latest/", None),
    "click-option-group": ("https://click-option-group.readthedocs.io/en/latest/", None),
}



"""
This plugin collects and runs the examples in :file:`doc/examples/`.

Examples need to have a :file:`test.console` file that contains shell commands
and their output similarly to Python doctests.
"""
import subprocess
from pathlib import Path
from typing import TYPE_CHECKING, Dict, Generator, List, Optional, Union

import pytest
from _pytest._code.code import TerminalRepr
from _pytest.assertion.util import _diff_text


if TYPE_CHECKING:
    from _pytest._code.code import ExceptionInfo, _TracebackStyle
    from _pytest._io import TerminalWriter
    from _pytest.config import Config
    from _pytest.main import Session


EXAMPLES_DIR = Path(__file__).parent.joinpath("examples")


# Part of pathlib only from py39
def is_relative_to(path: Path, other: Path) -> bool:
    """Return True if the path is relative to another path or False."""
    try:
        path.relative_to(other)
        return True
    except ValueError:
        return False


def pytest_ignore_collect(collection_path: Path, config: "Config") -> bool:
    """Do not collect Python files from the examples."""
    return (
        is_relative_to(collection_path, EXAMPLES_DIR)
        and collection_path.suffix == ".py"
    )


def pytest_collect_file(
    file_path: Path, parent: "Session"
) -> Optional["ExampleFile"]:
    """Checks if the file is a rst file and creates an
    :class:`ExampleFile` instance."""
    if (
        is_relative_to(file_path, EXAMPLES_DIR)
        and file_path.name == "test.console"
    ):
        return ExampleFile.from_parent(parent, path=file_path)
    return None


class ExampleFile(pytest.File):
    """Represents an example ``.py`` and its output ``.out``."""

    def collect(self) -> Generator["ExampleItem", None, None]:
        name = f"Example.{self.path.parent.name}"
        name = "console_session"
        yield ExampleItem.from_parent(self, name=name)


class ExampleItem(pytest.Item):
    """Executes an example found in a rst-file."""

    def runtest(self) -> None:
        # Read expected output.
        # The last line is an empty line, skip it.
        lines = self.path.read_text().splitlines()
        cmds: Dict[str, List[str]] = {}
        last_cmd = ""
        for line in lines:
            if line.startswith("$ "):
                last_cmd = line[2:]
                cmds[last_cmd] = []
            else:
                cmds[last_cmd].append(line)

        for cmd, output_lines in cmds.items():
            expected = "\n".join(output_lines)
            output = subprocess.run(
                cmd,
                shell=True,
                cwd=self.path.parent,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                universal_newlines=True,
                check=True,
            ).stdout
            # Normalize whitespace:
            output = "\n".join(l.rstrip() for l in output.splitlines())
            if output != expected:
                raise ValueError(cmd, output, expected)

    def repr_failure(
        self,
        excinfo: "ExceptionInfo[BaseException]",
        style: "Optional[_TracebackStyle]" = None,
    ) -> Union[TerminalRepr, str]:
        if excinfo.errisinstance(ValueError):
            # Output is mismatching. Create a nice diff as failure description.
            cmd, output, expected = excinfo.value.args
            diff_text = _diff_text(output, expected, 2)
            return ReprFailExample(self, cmd, diff_text)

        elif excinfo.errisinstance(subprocess.CalledProcessError):
            # Something went wrong while executing the example.
            return ReprErrorExample(self, excinfo)  # type: ignore

        # Something went terribly wrong :(
        return pytest.Item.repr_failure(self, excinfo)


class ReprFailExample(TerminalRepr):
    """Reports output mismatches in a nice and informative representation."""

    markup = {
        "+": dict(green=True),
        "-": dict(red=True),
        "?": dict(bold=True),
    }
    """Colorization codes for the diff markup."""

    def __init__(
        self, item: ExampleItem, cmd: str, diff_text: List[str]
    ) -> None:
        self.item = item
        self.cmd = cmd
        self.diff_text = diff_text

    def toterminal(self, tw: "TerminalWriter") -> None:
        tw.line()
        tw.line("Got unexpected output while running the console session:")
        tw.line()
        tw.line(f"$ {self.cmd}", bold=True)
        for line in self.diff_text:
            markup = self.markup.get(line[0], {})
            tw.line(line, **markup)
        tw.line()


class ReprErrorExample(TerminalRepr):
    """Reports failures in the execution of an example."""

    def __init__(
        self,
        item: ExampleItem,
        exc_info: "ExceptionInfo[subprocess.CalledProcessError]",
    ) -> None:
        self.item = item
        self.exc_info = exc_info

    def toterminal(self, tw: "TerminalWriter") -> None:
        exc = self.exc_info.value
        tw.line()
        tw.line("An error occurred while running the console session:")
        tw.line()
        tw.line(f"$ {exc.cmd}", bold=True)
        tw.line(self.exc_info.value.output)



"""
Extract examples from the README and assert they work.
"""
import pathlib
import subprocess
from typing import Dict, List, Tuple

import pytest


HERE = pathlib.Path(__file__).parent


def readme() -> str:
    """
    Returns the contents of the `README.md`.
    """
    return HERE.parent.joinpath("README.md").read_text(encoding="utf-8")


def load_readme() -> List[Tuple[str, List[str]]]:
    """
    Extracts the examples and returns a dict mapping example titles to a list
    of all lines in that section.
    """
    lines = readme().splitlines()
    in_examples = False
    example_title = None
    examples: Dict[str, List[str]] = {}
    for line in lines:
        if line == "## Examples":
            in_examples = True
            continue

        if not in_examples:
            continue

        if line.startswith("## "):
            return list(examples.items())

        if line.startswith("### "):
            example_title = line[4:]
            examples[example_title] = []
            continue

        if example_title:
            examples[example_title].append(line)

    raise AssertionError("We should not have gotten here")


@pytest.fixture
def example(request, tmp_path):
    """
    Splits the example lines into code blocks.

    If a code block starts with "# filename", write the contents to that file
    in *tmp_path*.

    Else, assume that the block contains a console session.  Lines starting
    with "$" are commands, other lines are the expected output of the above
    command.
    """
    example_lines = request.param
    code_lines = None
    for line in example_lines:
        if line.startswith("```") and len(line) > 3:
            code_lines = []
            continue

        if line == "```":
            if code_lines[0].startswith("# "):
                first_line, *code_lines = code_lines
                _, _, fname = first_line.partition(" ")
                contents = "\n".join(code_lines) + "\n"
                tmp_path.joinpath(fname).write_text(contents)
            else:
                cmds = {}
                current_cmd = None
                for line in code_lines:
                    if line.startswith("$"):
                        _, _, current_cmd = line.partition(" ")
                        cmds[current_cmd] = []
                    else:
                        cmds[current_cmd].append(line)
                return cmds
        elif code_lines is not None:
            code_lines.append(line)


@pytest.mark.parametrize(
    "example",
    [pytest.param(e[1], id=e[0]) for e in load_readme()],
    indirect=True,
)
def test_readme(example, tmp_path):
    """
    All commands in the *console* block of an example produce the exact same
    results as shown in the example.
    """
    for cmd, expected in example.items():
        result = subprocess.run(
            cmd,
            shell=True,
            cwd=str(tmp_path),
            capture_output=True,
            text=True,
            check=False,
        )
        assert result.stderr == ""
        assert result.stdout.splitlines() == expected
        assert result.returncode == 0



"""
Utilities for generating Click options
"""
import itertools
import typing as t
from collections.abc import (
    Mapping,
    MutableMapping,
    MutableSequence,
    MutableSet,
    Sequence,
)
from datetime import datetime
from enum import Enum
from functools import update_wrapper

import attrs
import cattrs
import click

from ._compat import get_args, get_origin
from ._core import T, _load_settings, default_loaders
from ._dict_utils import _deep_options, _get_path, _merge_dicts, _set_path
from .attrs import CLICK_KEY, METADATA_KEY, _SecretRepr
from .converters import default_converter, from_dict
from .loaders import Loader
from .types import OptionInfo, SettingsDict


try:
    from typing import Protocol
except ImportError:
    # Python 3.7
    from typing import _Protocol as Protocol  # type: ignore


CTX_KEY = "settings"


Callback = t.Callable[[click.Context, click.Option, t.Any], t.Any]
AnyFunc = t.Callable[..., t.Any]
Decorator = t.Callable[[AnyFunc], AnyFunc]
StrDict = t.Dict[str, t.Any]
TypeHandlerFunc = t.Callable[[type, t.Any], StrDict]


def click_options(
    cls: t.Type[T],
    loaders: t.Union[str, t.Sequence[Loader]],
    converter: t.Optional[cattrs.Converter] = None,
    type_handler: "t.Optional[TypeHandler]" = None,
    argname: t.Optional[str] = None,
    decorator_factory: "t.Optional[DecoratorFactory]" = None,
) -> t.Callable[[t.Callable], t.Callable]:
    """
    Generate :mod:`click` options for a CLI which override settins loaded via
    :func:`.load_settings()`.

    A single *cls* instance is passed to the decorated function -- by default
    as positional argument.

    Args:
        cls: Attrs class with options (and default values).

        loaders: Either a string with your app name or a list of settings
            :class:`Loader`'s.  If it is a string, use it with
            :func:`~typed_settings.default_loaders()` to get the defalt
            loaders.

        converter: An optional :class:`cattrs.Converter` used for converting
            option values to the required type.

            By default, :data:`typed_settings.attrs.converter` is used.

        type_handler: Helps creating proper click options for option types that
            are not natively supported by click.

        argname: An optional argument name.  If it is set, the settings
            instances is no longer passed as positional argument but as key
            word argument.

            This allows a function to be decorated with this function multiple
            times.

        decorator_factory: A class that generates Click decorators for options
            and settings classes.  This allows you to, e.g., use
            `option groups`_ via :class:`OptionGroupFactory`.  The default
            generates normal Click options via :class:`ClickOptionFactory`.

            .. _option groups: https://click-option-group.readthedocs.io

    Return:
        A decorator for a click command.

    Example:

      .. code-block:: python

         >>> import click
         >>> import typed_settings as ts
         >>>
         >>> @ts.settings
         ... class Settings: ...
         ...
         >>> @click.command()
         ... @ts.click_options(Settings, "example")
         ... def cli(settings):
         ...     print(settings)

    .. versionchanged:: 1.0.0
       Instead of a list of loaders, you can also just pass an application
       name.
    .. versionchanged:: 1.1.0
       Add the *argname* parameter.
    .. versionchanged:: 1.1.0
       Add the *decorator_factory* parameter.
    """
    cls = attrs.resolve_types(cls)
    options = _deep_options(cls)
    grouped_options = [
        (g_cls, list(g_opts))
        for g_cls, g_opts in itertools.groupby(options, key=lambda o: o.cls)
    ]

    if isinstance(loaders, str):
        loaders = default_loaders(loaders)

    settings_dict = _load_settings(cls, options, loaders)

    converter = converter or default_converter()
    type_handler = type_handler or TypeHandler()
    decorator_factory = decorator_factory or ClickOptionFactory()

    wrapper = _get_wrapper(
        cls,
        settings_dict,
        options,
        grouped_options,
        converter,
        type_handler,
        argname,
        decorator_factory,
    )
    return wrapper


def _get_wrapper(
    cls: t.Type[T],
    settings_dict: SettingsDict,
    options: t.List[OptionInfo],
    grouped_options: t.List[t.Tuple[type, t.List[OptionInfo]]],
    converter: cattrs.Converter,
    type_handler: "TypeHandler",
    argname: t.Optional[str],
    decorator_factory: "DecoratorFactory",
) -> t.Callable[[t.Callable], t.Callable]:
    def pass_settings(f: AnyFunc) -> Decorator:
        """
        Creates a *cls* instances from the settings dict stored in
        :attr:`click.Context.obj` and passes it to the decorated function *f*.
        """

        def new_func(*args: t.Any, **kwargs: t.Any) -> t.Any:
            ctx = click.get_current_context()
            if ctx.obj is None:
                ctx.obj = {}
            _merge_dicts(options, settings_dict, ctx.obj.get(CTX_KEY, {}))
            settings = from_dict(settings_dict, cls, converter)
            if argname:
                ctx_key = argname
                kwargs = {argname: settings, **kwargs}
            else:
                ctx_key = CTX_KEY
                args = (settings,) + args
            ctx.obj[ctx_key] = settings
            return f(*args, **kwargs)

        return update_wrapper(new_func, f)

    def wrap(f: AnyFunc) -> AnyFunc:
        """
        The wrapper that actually decorates a function with all options.
        """
        option_decorator = decorator_factory.get_option_decorator()
        for g_cls, g_opts in reversed(grouped_options):
            for oinfo in reversed(g_opts):
                default = _get_default(
                    oinfo.field, oinfo.path, settings_dict, converter
                )
                option = _mk_option(
                    option_decorator,
                    oinfo.path,
                    oinfo.field,
                    default,
                    type_handler,
                )
                f = option(f)
            f = decorator_factory.get_group_decorator(g_cls)(f)

        f = pass_settings(f)
        return f

    return wrap


def pass_settings(
    f: t.Optional[AnyFunc] = None, *, argname: t.Optional[str] = None
) -> AnyFunc:
    """
    Marks a callback as wanting to receive the innermost settings instance as
    first argument.

    If you specifiy an *argname* in :func:`click_options()`, you must specify
    the same name here in order to get the correct settings instance.  The
    settings instance is then passed as keyword argument.

    Args:
        argname: An optional argument name.  If it is set, the settings
            instances is no longer passed as positional argument but as key
            word argument.

    Return:
        A decorator for a click command.

    Example:

      .. code-block:: python

         >>> import click
         >>> import typed_settings as ts
         >>>
         >>> @ts.settings
         ... class Settings: ...
         ...
         >>> @click.group()
         ... @click_options(Settings, "example", argname="my_settings")
         ... def cli(my_settings):
         ...     pass
         ...
         >>> @cli.command()
         ... # Use the same "argname" as above!
         ... @pass_settings(argname="my_settings")
         ... def sub_cmd(*, my_settings):
         ...     print(my_settings)

    .. versionchanged:: 1.1.0
       Add the *argname* parameter.
    """
    ctx_key = argname or CTX_KEY

    def decorator(f: AnyFunc) -> AnyFunc:
        def new_func(*args: t.Any, **kwargs: t.Any) -> t.Any:
            ctx = click.get_current_context()
            node: t.Optional[click.Context] = ctx
            settings = None
            while node is not None:
                if isinstance(node.obj, dict) and ctx_key in node.obj:
                    settings = node.obj[ctx_key]
                    break
                node = node.parent

            if argname:
                kwargs = {argname: settings, **kwargs}
            else:
                args = (settings,) + args

            return ctx.invoke(f, *args, **kwargs)

        return update_wrapper(new_func, f)

    if f is None:
        return decorator

    return decorator(f)


class DecoratorFactory(Protocol):
    """
    **Protocol:** Methods that a Click decorator factory must implement.

    The decorators returned by the procol methods are used to construct the
    Click options and possibly option groups.

    .. versionadded:: 1.1.0
    """

    def get_option_decorator(self) -> t.Callable[..., Decorator]:
        """
        Return the decorator that is used for creating Click options.

        It must be compatible with :func:`click.option()`.
        """
        ...

    def get_group_decorator(self, settings_cls: type) -> Decorator:
        """
        Return a decorator for the current settings class.

        This can, e.g., be used to group option by settings class.
        """
        ...


class ClickOptionFactory:
    """
    Factory for default Click decorators.
    """

    def get_option_decorator(self) -> t.Callable[..., Decorator]:
        """
        Return :func:`click.option()`.
        """
        return click.option

    def get_group_decorator(self, settings_cls: type) -> Decorator:
        """
        Return a no-op decorator that leaves the decorated function unchanged.
        """
        return lambda f: f


class OptionGroupFactory:
    """
    Factory got generating Click option groups via
    https://click-option-group.readthedocs.io.
    """

    def __init__(self) -> None:
        try:
            from click_option_group import optgroup
        except ImportError as e:
            raise ModuleNotFoundError(
                "Module 'click_option_group' not installed.  "
                "Please run 'python -m pip install click-option-group'"
            ) from e
        self.optgroup = optgroup

    def get_option_decorator(self) -> t.Callable[..., Decorator]:
        """
        Return :func:`click_option_group.optgroup.option()`.
        """
        return self.optgroup.option

    def get_group_decorator(self, settings_cls: type) -> Decorator:
        """
        Return a :func:`click_option_group.optgroup.group()` instantiated with
        the first line of *settings_cls*'s docstring.
        """
        try:
            name = settings_cls.__doc__.strip().splitlines()[0]  # type: ignore
        except (AttributeError, IndexError):
            name = f"{settings_cls.__name__} options"
        return self.optgroup.group(name)


def handle_datetime(type: type, default: t.Any) -> StrDict:
    """
    Use :class:`click.DateTime` as option type and convert the default value
    to an ISO string.
    """
    type_info = {
        "type": click.DateTime(
            ["%Y-%m-%d", "%Y-%m-%dT%H:%M:%S", "%Y-%m-%dT%H:%M:%S%z"]
        ),
    }
    if default is not attrs.NOTHING:
        type_info["default"] = default.isoformat()
    return type_info


def handle_enum(type: t.Type[Enum], default: t.Any) -> StrDict:
    """
    Use :class:`EnumChoice` as option type and use the enum value's name as
    default.
    """
    type_info = {"type": click.Choice(list(type.__members__))}
    if default is not attrs.NOTHING:
        # Convert Enum instance to string
        type_info["default"] = default.name

    return type_info


#: Default handlers for click option types.
DEFAULT_TYPES: t.Dict[type, TypeHandlerFunc] = {
    datetime: handle_datetime,
    Enum: handle_enum,
}


class TypeHandler:
    """
    This class derives type information for Click options from an Attrs field's
    type.

    The class differentitates between specific and generic types (e.g.,
    :samp:`int` vs. :samp:`List[{T}]`.

    Specific types:
        Handlers for specific types can be extended and modified by passing
        a *types* dict to the class.  By default, :data:`DEFAULT_TYPES` is
        used.

        This dict maps Python types to a handler function.  Handler functions
        take the field type and default value and return a dict that is passed
        as keyword arguments to :func:`click.option()`.  This dict should
        contain a ``type`` key and, optionally, an updated ``default``.

        .. code-block:: python

            def handle_mytype(type: type, default: Any) -> t.Dict[str, Any]:
                type_info = {
                    "type": ClickType(...)
                }
                if default is not attrs.NOTHING:
                    type_info["default"] = default.stringify()
                return type_info

        You can use :func:`handle_datetime` and :func:`handle_enum` as
        a sample.

        t.Types without a handler get no special treatment and cause options to
        look like this: :samp:`click.option(..., type=field_type,
        default=field_default)`.

    Generic types:
        Handlers for generic types cannot be changed.  They either create an
        option with :samp:`multiple=True` or :samp:`nargs={x}`.  Nested types
        are recursively resolved.

        t.Types that cause :samp:`multiple=True`:

        - :class:`typing.List`
        - :class:`typing.Sequence`
        - :class:`typing.MutableSequence`
        - :class:`typing.Set`
        - :class:`typing.FrozenSet`
        - :class:`typing.MutableSet`

        t.Types that cause :samp:`nargs={x}`:

        - :class:`typing.Tuple`
        - :class:`typing.NamedTuple`

        Dicts are not (yet) supported.
    """

    def __init__(
        self,
        types: t.Optional[t.Dict[type, TypeHandlerFunc]] = None,
    ) -> None:
        self.types = types or DEFAULT_TYPES
        self.list_types = (
            list,
            Sequence,
            MutableSequence,
            set,
            frozenset,
            MutableSet,
        )
        self.tuple_types = (tuple,)
        self.mapping_types = (
            dict,
            Mapping,
            MutableMapping,
        )

    def get_type(self, otype: t.Optional[type], default: t.Any) -> StrDict:
        """
        Analyses the option type and returns updated options.
        """
        origin = get_origin(otype)
        args = get_args(otype)

        if otype is None:
            return self._handle_basic_types(otype, default)

        elif origin is None:
            for target_type, get_type_info in self.types.items():
                if issubclass(otype, target_type):
                    return get_type_info(otype, default)

            return self._handle_basic_types(otype, default)

        else:
            if origin in self.list_types:
                return self._handle_list(otype, default, args)
            elif origin in self.tuple_types:
                return self._handle_tuple(otype, default, args)
            elif origin in self.mapping_types:
                return self._handle_dict(otype, default, args)

            raise TypeError(f"Cannot create click type for: {otype}")

    def _handle_basic_types(
        self, type: t.Optional[type], default: t.Any
    ) -> StrDict:
        if default is attrs.NOTHING:
            type_info = {"type": type}
        else:
            type_info = {"type": type, "default": default}
        return type_info

    def _handle_list(
        self, type: type, default: t.Any, args: t.Tuple[t.Any, ...]
    ) -> StrDict:
        # lists and list-like tuple
        type_info = self.get_type(args[0], attrs.NOTHING)
        if default is not attrs.NOTHING:
            default = [self.get_type(args[0], d)["default"] for d in default]
            type_info["default"] = default
        type_info["multiple"] = True
        return type_info

    def _handle_tuple(
        self, type: type, default: t.Any, args: t.Tuple[t.Any, ...]
    ) -> StrDict:
        if len(args) == 2 and args[1] == ...:
            return self._handle_list(type, default, args)
        else:
            # "struct" variant of tuple
            if default is attrs.NOTHING:
                default = [attrs.NOTHING] * len(args)
            dicts = [self.get_type(a, d) for a, d in zip(args, default)]
            type_info = {
                "type": tuple(d["type"] for d in dicts),
                "nargs": len(dicts),
            }
            if all("default" in d for d in dicts):
                type_info["default"] = tuple(d["default"] for d in dicts)
            return type_info

    def _handle_dict(
        self, type: type, default: t.Any, args: t.Tuple[t.Any, ...]
    ) -> StrDict:
        def cb(
            ctx: click.Context,
            param: click.Option,
            value: t.Optional[t.Iterable[str]],
        ) -> t.Dict[str, str]:
            if not value:
                return {}
            splitted = [v.partition("=") for v in value]
            items = {k: v for k, _, v in splitted}
            return items

        type_info = {
            "metavar": "KEY=VALUE",
            "multiple": True,
            "callback": cb,
        }
        if default is not attrs.NOTHING:
            default = [f"{k}={v}" for k, v in default.items()]
            type_info["default"] = default
        return type_info


def _get_default(
    field: attrs.Attribute,
    path: str,
    settings: SettingsDict,
    converter: cattrs.Converter,
) -> t.Any:
    """
    Returns the proper default value for an attribute.

    If possible, the default is taken from loaded settings.  Else, use the
    field's default value.
    """
    try:
        # Use loaded settings value
        default = _get_path(settings, path)
    except KeyError:
        # Use field's default
        default = field.default
    else:
        # If the default was found (no KeyError), convert the input value to
        # the proper type.
        # See: https://gitlab.com/sscherfke/typed-settings/-/issues/11
        if field.type:
            default = converter.structure(default, field.type)

    if isinstance(default, attrs.Factory):  # type: ignore
        if default.takes_self:
            # There is no instance yet.  Passing ``None`` migh be more correct
            # than passing a fake instance, because it raises an error instead
            # of silently creating a false value. :-?
            default = default.factory(None)
        else:
            default = default.factory()

    return default


def _mk_option(
    option: t.Callable[..., Decorator],
    path: str,
    field: attrs.Attribute,
    default: t.Any,
    type_handler: TypeHandler,
) -> Decorator:
    """
    Recursively creates click options and returns them as a list.
    """
    user_config = field.metadata.get(METADATA_KEY, {}).get(CLICK_KEY, {})

    param_decls: t.Tuple[str, ...]
    user_param_decls: t.Union[str, t.Sequence[str]]
    user_param_decls = user_config.pop("param_decls", ())
    if not user_param_decls:
        option_name = path.replace(".", "-").replace("_", "-")
        if field.type and field.type is bool:
            param_decls = (f"--{option_name}/--no-{option_name}",)
        else:
            param_decls = (f"--{option_name}",)
    elif isinstance(user_param_decls, str):
        param_decls = (user_param_decls,)
    else:
        param_decls = tuple(user_param_decls)

    # The option type specifies the default option kwargs
    kwargs = type_handler.get_type(field.type, default)

    # The type's kwargs should not be able to set these values since they are
    # needed for everything to work:
    kwargs["show_default"] = True
    kwargs["expose_value"] = False
    kwargs["callback"] = _make_callback(path, kwargs.get("callback"))

    # Get "help" from the user_config *now*, because we may need to update it
    # below.  Also replace "None" with "".
    kwargs["help"] = user_config.pop("help", None) or ""

    if isinstance(field.repr, _SecretRepr):
        kwargs["show_default"] = False
        if default is not attrs.NOTHING:  # pragma: no cover
            kwargs["help"] = f"{kwargs['help']}  [default: {field.repr('')}]"

    if default is attrs.NOTHING:
        kwargs["required"] = True

    # The user has the last word, though.
    kwargs.update(user_config)

    return option(*param_decls, **kwargs)


def _make_callback(path: str, type_callback: t.Optional[Callback]) -> Callback:
    """
    Generate a callback that adds option values to the settings instance in the
    context.  It also calls a type's callback if there should be one.
    """

    def cb(ctx: click.Context, param: click.Option, value: t.Any) -> t.Any:
        if type_callback is not None:
            value = type_callback(ctx, param, value)

        if ctx.obj is None:
            ctx.obj = {}
        settings = ctx.obj.setdefault(CTX_KEY, {})
        _set_path(settings, path, value)
        return value

    return cb



import pytest

from typed_settings._dict_utils import _deep_options
from typed_settings.attrs import option, settings
from typed_settings.types import OptionList


# Test with frozen settings.  If it works this way, it will also work with
# mutable settings but not necessarily the other way around.
@settings(frozen=True)
class Host:
    name: str
    port: int = option(converter=int)


@settings(frozen=True)
class Settings:
    host: Host
    url: str
    default: int = 3


@pytest.fixture
def settings_cls() -> type:
    return Settings


@pytest.fixture
def options(settings_cls: type) -> OptionList:
    return _deep_options(settings_cls)



"""
Tests for `typed_settings.attrs.converters`.
"""
import json
import sys
from datetime import datetime, timedelta, timezone
from enum import Enum
from itertools import product
from pathlib import Path
from typing import Any, Dict, FrozenSet, List, Optional, Sequence, Set, Tuple

import pytest

from typed_settings.attrs import option, secret, settings
from typed_settings.converters import (
    default_converter,
    from_dict,
    register_strlist_hook,
    to_bool,
    to_dt,
    to_enum,
    to_path,
)
from typed_settings.exceptions import InvalidValueError


class LeEnum(Enum):
    spam = "Le Spam"
    eggs = "Le Eggs"


@settings
class S:
    u: str = option()
    p: str = secret()


class TestToDt:
    """Tests for `to_dt`."""

    def test_from_dt(self):
        """
        Existing datetimes are returned unchanged.
        """
        dt = datetime(2020, 5, 4, 13, 37)
        result = to_dt(dt, datetime)
        assert result is dt

    @pytest.mark.parametrize(
        "value, expected",
        [
            ("2020-05-04 13:37:00", datetime(2020, 5, 4, 13, 37)),
            ("2020-05-04T13:37:00", datetime(2020, 5, 4, 13, 37)),
            (
                "2020-05-04T13:37:00Z",
                datetime(2020, 5, 4, 13, 37, tzinfo=timezone.utc),
            ),
            (
                "2020-05-04T13:37:00+00:00",
                datetime(2020, 5, 4, 13, 37, tzinfo=timezone.utc),
            ),
            (
                "2020-05-04T13:37:00+02:00",
                datetime(
                    2020,
                    5,
                    4,
                    13,
                    37,
                    tzinfo=timezone(timedelta(seconds=7200)),
                ),
            ),
        ],
    )
    def test_from_str(self, value, expected):
        """
        Existing datetimes are returned unchanged.
        """
        result = to_dt(value, datetime)
        assert result == expected

    def test_invalid_input(self):
        """
        Invalid inputs raises a TypeError.
        """
        with pytest.raises(TypeError):
            to_dt(3)


class TestToBool:
    """Tests for `to_bool`."""

    @pytest.mark.parametrize(
        "val, expected",
        [
            (True, True),
            ("True", True),
            ("TRUE", True),
            ("true", True),
            ("t", True),
            ("yes", True),
            ("Y", True),
            ("on", True),
            ("1", True),
            (1, True),
            (False, False),
            ("False", False),
            ("false", False),
            ("fAlse", False),
            ("NO", False),
            ("n", False),
            ("off", False),
            ("0", False),
            (0, False),
        ],
    )
    def test_to_bool(self, val, expected):
        """
        Only a limited set of values can be converted to a bool.
        """
        assert to_bool(val, bool) is expected

    @pytest.mark.parametrize("val", ["", [], "spam", 2, -1])
    def test_to_bool_error(self, val):
        """
        In contrast to ``bool()``, `to_bool` does no take Pythons default
        truthyness into account.

        Everything that is not in the sets above raises an error.
        """
        pytest.raises(ValueError, to_bool, val, bool)


class TestToEnum:
    """Tests for `to_enum`."""

    @pytest.mark.parametrize(
        "value, expected",
        [
            (LeEnum.spam, LeEnum.spam),
            ("spam", LeEnum.spam),
        ],
    )
    def test_to_enum(self, value, expected):
        """
        `to_enum()` accepts Enums and member names.
        """
        assert to_enum(value, LeEnum) is expected


class TestToPath:
    """Tests for `to_path`."""

    @pytest.mark.parametrize(
        "value, expected",
        [
            ("spam", Path("spam")),
            (Path("eggs"), Path("eggs")),
        ],
    )
    def test_to_path(self, value, expected):
        assert to_path(value, Path) == expected


@pytest.mark.parametrize(
    "typ, value, expected",
    [
        # Bools can be parsed from a defined set of values
        (bool, True, True),
        (bool, "True", True),
        (bool, "true", True),
        (bool, "yes", True),
        (bool, "1", True),
        (bool, 1, True),
        (bool, False, False),
        (bool, "False", False),
        (bool, "false", False),
        (bool, "no", False),
        (bool, "0", False),
        (bool, 0, False),
        # Other simple types
        (int, 23, 23),
        (int, "42", 42),
        (float, 3.14, 3.14),
        (float, ".815", 0.815),
        (str, "spam", "spam"),
        (datetime, "2020-05-04T13:37:00", datetime(2020, 5, 4, 13, 37)),
        # Enums are parsed from their "key"
        (LeEnum, "eggs", LeEnum.eggs),
        # (Nested) attrs classes
        (S, {"u": "user", "p": "pwd"}, S("user", "pwd")),
        (S, S("user", "pwd"), S("user", "pwd")),
        # Container types
        (List[int], [1, 2], [1, 2]),
        (List[S], [{"u": 1, "p": 2}], [S("1", "2")]),
        (Dict[str, int], {"a": 1, "b": 3.1}, {"a": 1, "b": 3}),
        (Dict[str, S], {"a": {"u": "u", "p": "p"}}, {"a": S("u", "p")}),
        (Tuple[str, ...], [1, "2", 3], ("1", "2", "3")),
        (Tuple[int, bool, str], [0, "0", 0], (0, False, "0")),
        # "Special types"
        (Any, 2, 2),
        (Any, "2", "2"),
        (Any, None, None),
        (Optional[str], 1, "1"),
        (Optional[S], None, None),
        (Optional[S], {"u": "u", "p": "p"}, S("u", "p")),
        (Optional[LeEnum], "spam", LeEnum.spam),
    ],
)
def test_supported_types(typ, value, expected):
    """
    All oficially supported types can be converted by attrs.

    Please create an issue if something is missing here.
    """

    @settings
    class Settings:
        opt: typ

    inst = from_dict({"opt": value}, Settings, default_converter())
    assert inst.opt == expected


@pytest.mark.parametrize("val", [{"foo": 3}, {"opt", "x"}])
def test_unsupported_values(val):
    """
    An InvalidValueError is raised if a settings dict cannot be converted
    to the settings class.
    """

    @settings
    class Settings:
        opt: int

    with pytest.raises(InvalidValueError):
        from_dict(val, Settings, default_converter())


STRLIST_TEST_DATA = [
    (List[int], [1, 2, 3]),
    (Sequence[int], [1, 2, 3]),
    (Set[int], {1, 2, 3}),
    (FrozenSet[int], frozenset({1, 2, 3})),
    (Tuple[int, ...], (1, 2, 3)),
    (Tuple[int, int, int], (1, 2, 3)),
]

if sys.version_info[:2] >= (3, 9):
    STRLIST_TEST_DATA.extend(
        [
            (list[int], [1, 2, 3]),
            (set[int], {1, 2, 3}),
            (tuple[int, ...], (1, 2, 3)),  # type: ignore
        ]
    )


@pytest.mark.parametrize(
    "input, kw, typ, expected",
    [
        (input, kw, typ, expected)
        for (input, kw), (typ, expected) in product(
            [("1:2:3", {"sep": ":"}), ("[1,2,3]", {"fn": json.loads})],
            STRLIST_TEST_DATA,
        )
    ],
)
def test_strlist_hook(input, kw, typ, expected):
    @settings
    class Settings:
        a: typ

    converter = default_converter()
    register_strlist_hook(converter, **kw)
    inst = from_dict({"a": input}, Settings, converter)
    assert inst == Settings(expected)


def test_strlist_hook_either_arg():
    """
    Either "sep" OR "fn" can be passed to "register_str_list_hook()"
    """
    converter = default_converter()
    with pytest.raises(ValueError, match="You may either pass"):
        register_strlist_hook(converter, sep=":", fn=lambda v: [v])



from typed_settings import types


def test_auto_singleton():
    """
    `_Auto()`  is a singleton.
    """
    assert types._Auto() is types.AUTO


def test_auto_repr():
    """
    `_Auto()` has a nice repr.
    """
    assert repr(types._Auto()) == "AUTO"



import sys


if sys.version_info[:2] >= (3, 8):
    from typing import get_args, get_origin
else:
    import collections.abc
    from typing import Generic, _GenericAlias

    def get_origin(tp):
        # Backported from py38
        if isinstance(tp, _GenericAlias):
            return tp.__origin__
        if tp is Generic:  # pragma: no cover
            return Generic
        return None

    def get_args(tp):
        # Backported from py38
        if isinstance(tp, _GenericAlias) and not tp._special:
            res = tp.__args__
            if (  # pragma: no cover
                get_origin(tp) is collections.abc.Callable
                and res[0] is not Ellipsis  # noqa: W503
            ):
                res = (list(res[:-1]), res[-1])  # pragma: no cover
            return res
        return ()


__all__ = ["get_args", "get_origin"]



from typing import Tuple

import click
import typed_settings as ts
from typed_settings.click_utils import OptionGroupFactory


@ts.settings
class Coverage:
    """
    Coverage settings
    """

    src: str = ts.option(default="", click={"param_decls": ("--cov", "src")})
    report: str = ""


@ts.settings
class Emoji:
    """
    Emoji settings
    """

    enable: bool = True


@ts.settings
class Base:
    """
    Main settings
    """

    marker: str = ts.option(
        default="",
        help="only run tests which macht the given substring expression",
        click={"param_decls": ("-m",)},
    )
    exitfirst: bool = ts.option(
        default=False,
        help="Exit instantly on first error or failed test",
        click={"param_decls": ("--exitfirst", "-x"), "is_flag": True},
    )
    stepwise: bool = ts.option(
        default=False,
        help=(
            "Exit on test failure and continue from last failing test next "
            "time"
        ),
        click={"param_decls": ("--stepwise", "--sw"), "is_flag": True},
    )


Settings = ts.combine(
    "Settings",
    Base,
    {
        # Imagine, this dict comes from a "load_plugins()" function :)
        "cov": Coverage(),
        "emoji": Emoji(),
    },
)


@click.command()
@click.argument("file_or_dir", nargs=-1)
@ts.click_options(Settings, "pytest", decorator_factory=OptionGroupFactory())
def cli(settings: Settings, file_or_dir: Tuple[str, ...]):
    print(settings)


if __name__ == "__main__":
    cli()



"""
Helpers for and additions to :mod:`attrs`.
"""
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Mapping,
    Optional,
    overload,
)

import attr  # The old namespaces is needed in "combine()"
import attrs


if TYPE_CHECKING:
    try:
        from attr import (  # type: ignore
            _T,
            _ConverterType,
            _OnSetAttrArgType,
            _ReprArgType,
            _ValidatorArgType,
        )
    except ImportError:
        # Just in case the symbols are moved from "attr" to "attrs"
        from attrs import (  # type: ignore
            _T,
            _ConverterType,
            _OnSetAttrArgType,
            _ReprArgType,
            _ValidatorArgType,
        )

from .hooks import auto_convert


__all__ = [
    "METADATA_KEY",
    "SECRET",
    "auto_convert",
    "evolve",
    "option",
    "secret",
    "settings",
]


METADATA_KEY = "typed_settings"
CLICK_KEY = "click"


class _SecretRepr:
    def __call__(self, _v: Any) -> str:
        return "***"

    def __repr__(self) -> str:
        return "***"


SECRET = _SecretRepr()


settings = attrs.define
"""An alias to :func:`attrs.define()`"""


@overload
def option(
    *,
    default: None = ...,
    validator: None = ...,
    repr: "_ReprArgType" = ...,
    hash: Optional[bool] = ...,
    init: bool = ...,
    metadata: Optional[Dict[Any, Any]] = ...,
    converter: None = ...,
    factory: None = ...,
    kw_only: bool = ...,
    eq: Optional[bool] = ...,
    order: Optional[bool] = ...,
    on_setattr: Optional["_OnSetAttrArgType"] = ...,
    help: Optional[str] = ...,
    click: Optional[Dict[str, Any]] = ...,
) -> Any:
    ...


# This form catches an explicit None or no default and infers the type from the
# other arguments.
@overload
def option(
    *,
    default: None = ...,
    validator: "Optional[_ValidatorArgType[_T]]" = ...,
    repr: "_ReprArgType" = ...,
    hash: Optional[bool] = ...,
    init: bool = ...,
    metadata: Optional[Dict[Any, Any]] = ...,
    converter: Optional["_ConverterType"] = ...,
    factory: "Optional[Callable[[], _T]]" = ...,
    kw_only: bool = ...,
    eq: Optional[bool] = ...,
    order: Optional[bool] = ...,
    on_setattr: "Optional[_OnSetAttrArgType]" = ...,
    help: Optional[str] = ...,
    click: Optional[Dict[str, Any]] = ...,
) -> "_T":
    ...


# This form catches an explicit default argument.
@overload
def option(
    *,
    default: "_T",
    validator: "Optional[_ValidatorArgType[_T]]" = ...,
    repr: "_ReprArgType" = ...,
    hash: Optional[bool] = ...,
    init: bool = ...,
    metadata: Optional[Dict[Any, Any]] = ...,
    converter: "Optional[_ConverterType]" = ...,
    factory: "Optional[Callable[[], _T]]" = ...,
    kw_only: bool = ...,
    eq: Optional[bool] = ...,
    order: Optional[bool] = ...,
    on_setattr: "Optional[_OnSetAttrArgType]" = ...,
    help: Optional[str] = ...,
    click: Optional[Dict[str, Any]] = ...,
) -> "_T":
    ...


# This form covers type=non-Type: e.g. forward references (str), Any
@overload
def option(
    *,
    default: Optional["_T"] = ...,
    validator: "Optional[_ValidatorArgType[_T]]" = ...,
    repr: "_ReprArgType" = ...,
    hash: Optional[bool] = ...,
    init: bool = ...,
    metadata: Optional[Dict[Any, Any]] = ...,
    converter: "Optional[_ConverterType]" = ...,
    factory: "Optional[Callable[[], _T]]" = ...,
    kw_only: bool = ...,
    eq: Optional[bool] = ...,
    order: Optional[bool] = ...,
    on_setattr: "Optional[_OnSetAttrArgType]" = ...,
    help: Optional[str] = ...,
    click: Optional[Dict[str, Any]] = ...,
) -> Any:
    ...


def option(  # type: ignore[no-untyped-def]
    *,
    default=attrs.NOTHING,
    validator=None,
    repr=True,
    hash=None,
    init=True,
    metadata=None,
    converter=None,
    factory=None,
    kw_only=False,
    eq=None,
    order=None,
    on_setattr=None,
    help=None,
    click=None,
):
    """An alias to :func:`attrs.field()`"""
    metadata = _get_metadata(metadata, help, click)

    return attrs.field(
        default=default,
        validator=validator,
        repr=repr,
        hash=hash,
        init=init,
        metadata=metadata,
        converter=converter,
        factory=factory,
        kw_only=kw_only,
        eq=eq,
        order=order,
        on_setattr=on_setattr,
    )


@overload
def secret(
    *,
    default: None = ...,
    validator: None = ...,
    repr: _SecretRepr = ...,
    hash: Optional[bool] = ...,
    init: bool = ...,
    metadata: Optional[Dict[Any, Any]] = ...,
    converter: None = ...,
    factory: None = ...,
    kw_only: bool = ...,
    eq: Optional[bool] = ...,
    order: Optional[bool] = ...,
    on_setattr: "Optional[_OnSetAttrArgType]" = ...,
    help: Optional[str] = ...,
    click: Optional[Dict[str, Any]] = ...,
) -> Any:
    ...


# This form catches an explicit None or no default and infers the type from the
# other arguments.
@overload
def secret(
    *,
    default: None = ...,
    validator: "Optional[_ValidatorArgType[_T]]" = ...,
    repr: _SecretRepr = ...,
    hash: Optional[bool] = ...,
    init: bool = ...,
    metadata: Optional[Dict[Any, Any]] = ...,
    converter: "Optional[_ConverterType]" = ...,
    factory: "Optional[Callable[[], _T]]" = ...,
    kw_only: bool = ...,
    eq: Optional[bool] = ...,
    order: Optional[bool] = ...,
    on_setattr: "Optional[_OnSetAttrArgType]" = ...,
    help: Optional[str] = ...,
    click: Optional[Dict[str, Any]] = ...,
) -> "_T":
    ...


# This form catches an explicit default argument.
@overload
def secret(
    *,
    default: "_T",
    validator: "Optional[_ValidatorArgType[_T]]" = ...,
    repr: _SecretRepr = ...,
    hash: Optional[bool] = ...,
    init: bool = ...,
    metadata: Optional[Dict[Any, Any]] = ...,
    converter: "Optional[_ConverterType]" = ...,
    factory: "Optional[Callable[[], _T]]" = ...,
    kw_only: bool = ...,
    eq: Optional[bool] = ...,
    order: Optional[bool] = ...,
    on_setattr: "Optional[_OnSetAttrArgType]" = ...,
    help: Optional[str] = ...,
    click: Optional[Dict[str, Any]] = ...,
) -> "_T":
    ...


# This form covers type=non-Type: e.g. forward references (str), Any
@overload
def secret(
    *,
    default: "Optional[_T]" = ...,
    validator: "Optional[_ValidatorArgType[_T]]" = ...,
    repr: _SecretRepr = ...,
    hash: Optional[bool] = ...,
    init: bool = ...,
    metadata: Optional[Dict[Any, Any]] = ...,
    converter: "Optional[_ConverterType]" = ...,
    factory: "Optional[Callable[[], _T]]" = ...,
    kw_only: bool = ...,
    eq: Optional[bool] = ...,
    order: Optional[bool] = ...,
    on_setattr: "Optional[_OnSetAttrArgType]" = ...,
    help: Optional[str] = ...,
    click: Optional[Dict[str, Any]] = ...,
) -> Any:
    ...


def secret(  # type: ignore[no-untyped-def]
    *,
    default=attrs.NOTHING,
    validator=None,
    repr=SECRET,
    hash=None,
    init=True,
    metadata=None,
    converter=None,
    factory=None,
    kw_only=False,
    eq=None,
    order=None,
    on_setattr=None,
    help=None,
    click=None,
):
    """
    An alias to :func:`option()` but with a default repr that hides screts.

    When printing a settings instances, secret settings will represented with
    `***` istead of their actual value.

    See also:

        All arguments are describted here:

        - :func:`option()`
        - :func:`attrs.field()`

    Example:

    .. code-block:: python

        >>> from typed_settings import settings, secret
        >>>
        >>> @settings
        ... class Settings:
        ...     password: str = secret()
        ...
        >>> Settings(password="1234")
        Settings(password=***)
    """
    metadata = _get_metadata(metadata, help, click)

    return attrs.field(
        default=default,
        validator=validator,
        repr=repr,
        hash=hash,
        init=init,
        metadata=metadata,
        converter=converter,
        factory=factory,
        kw_only=kw_only,
        eq=eq,
        order=order,
        on_setattr=on_setattr,
    )


def _get_metadata(
    metadata: Optional[Dict[str, Any]], help: str, click: Dict[str, Any]
) -> Dict[str, Any]:
    click_config = {"help": help}
    if click:
        click_config.update(click)
    if metadata is None:
        metadata = {}
    ts_meta = metadata.setdefault(METADATA_KEY, {})
    ts_meta["help"] = help
    ts_meta[CLICK_KEY] = click_config
    return metadata


def evolve(inst: "_T", **changes: Any) -> "_T":
    """
    Create a new instance, based on *inst* with *changes* applied.

    If the old value of an attribute is an ``attrs`` class and the new value
    is a dict, the old value is updated recursively.

    .. warning::

       This function is very similar to :func:`attrs.evolve()`, but the
       ``attrs`` version is not updating values recursively.  Instead, it will
       just replace ``attrs`` instances with a dict.

    Args:
        inst: Instance of a class with ``attrs`` attributes.
        changes: Keyword changes in the new copy.

    Return:
        A copy of *inst* with *changes* incorporated.

    Raise:
        TypeError: If *attr_name* couldn't be found in the class ``__init__``.
        attrs.exceptions.NotAnAttrsClassError: If *cls* is not an ``attrs``
        class.

    ..  versionadded:: 1.0.0
    """
    cls = inst.__class__
    attribs = attrs.fields(cls)
    for a in attribs:
        if not a.init:
            continue
        attr_name = a.name  # To deal with private attributes.
        init_name = attr_name if attr_name[0] != "_" else attr_name[1:]
        old_value = getattr(inst, attr_name)
        if init_name not in changes:
            # Add original value to changes
            changes[init_name] = old_value
        elif attrs.has(old_value) and isinstance(changes[init_name], Mapping):
            # Evolve nested attrs classes
            changes[init_name] = evolve(old_value, **changes[init_name])

    return cls(**changes)


def combine(name: str, base_cls: type, nested: Dict[str, object]) -> type:
    """
    Create a new class called *name* based on *base_class* with additional
    attributes for *nested* classes.

    The same effect can be achieved by manually composing settings classes.
    A use case for this method is to combine settings classes from dynamically
    loaded plugins with the base settings of the main program.

    Args:
        name: The name for the new class.
        base_cls: The base class from which to copy all attributes.
        nested: A mapping of attribute names to (settings) class instances
            for which to generated additional attributes.  The attribute's
            type is the instance's type and its default value is the instance
            itself.  Keys in this dict must not overlap with the attributes
            of *base_cls*.

    Return:
        The created class *name*.

    Raise:
        ValueError: If *nested* contains a key for which *base_cls* already
            defines an attribute.

    Example:

        .. code-block:: python

           >>> import typed_settings as ts
           >>>
           >>> @ts.settings
           ... class Nested1:
           ...     a: str = ""
           >>>
           >>> @ts.settings
           ... class Nested2:
           ...     a: str = ""
           >>>
           >>> # Static composition
           >>> @ts.settings
           ... class Composed1:
           ...     a: str = ""
           ...     n1: Nested1 = Nested1()
           ...     n2: Nested2 = Nested2()
           ...
           >>> Composed1()
           Composed1(a='', n1=Nested1(a=''), n2=Nested2(a=''))
           >>>
           >>> # Dynamic composition
           >>> @ts.settings
           ... class BaseSettings:
           ...     a: str = ""
           >>>
           >>> Composed2 = ts.combine(
           ...     "Composed2",
           ...     BaseSettings,
           ...     {"n1": Nested1(), "n2": Nested2()},
           ... )
           >>> Composed2()
           Composed2(a='', n1=Nested1(a=''), n2=Nested2(a=''))


    .. versionadded:: 1.1.0
    """

    attribs = {
        a.name: attr.attrib(  # type: ignore[misc]
            default=a.default,
            validator=a.validator,  # type: ignore[arg-type]
            repr=a.repr,
            hash=a.hash,
            init=a.init,
            metadata=a.metadata,
            type=a.type,  # type: ignore[arg-type]
            converter=a.converter,  # type: ignore[arg-type]
            kw_only=a.kw_only,
            eq=a.eq,
            order=a.order,
            on_setattr=a.on_setattr,  # type: ignore[arg-type]
        )
        for a in attr.fields(base_cls)
    }
    for aname, default in nested.items():
        if aname in attribs:
            raise ValueError(f"Duplicate attribute for nested class: {aname}")
        attribs[aname] = attr.attrib(default=default, type=default.__class__)

    cls = attr.make_class(name, attribs)
    cls.__doc__ = base_cls.__doc__
    return cls



"""
Internal data structures.
"""
from enum import Enum
from typing import Any, List, MutableMapping, TypeVar

import attrs


T = TypeVar("T")
ET = TypeVar("ET", bound=Enum)
SettingsDict = MutableMapping[str, Any]


class _Auto:
    """
    Sentinel class to indicate the lack of a value when ``None`` is ambiguous.

    ``_Auto`` is a singleton. There is only ever one of it.
    """

    _singleton = None

    def __new__(cls) -> "_Auto":
        if _Auto._singleton is None:
            _Auto._singleton = super(_Auto, cls).__new__(cls)
        return _Auto._singleton

    def __repr__(self) -> str:
        return "AUTO"


AUTO = _Auto()
"""
Sentinel to indicate the lack of a value when ``None`` is ambiguous.
"""


@attrs.frozen
class OptionInfo:
    """
    Information about (possibly nested) option attributes.

    Each instance represents a single attribute of an apps's settings class.
    """

    path: str
    """
    Dotted path to the option name relative to the root settings class.
    """

    field: attrs.Attribute
    """
    :class:`attrs.Attribute` instance for the option.
    """

    cls: type
    """
    The option's settings class.  This is either the root settings class or a
    nested one.
    """


OptionList = List[OptionInfo]



import textwrap
from itertools import product
from pathlib import Path
from typing import Any, Dict, List, Optional

import pytest
from pytest import MonkeyPatch

from typed_settings._dict_utils import _deep_options
from typed_settings.attrs import settings
from typed_settings.exceptions import (
    ConfigFileLoadError,
    ConfigFileNotFoundError,
    InvalidOptionsError,
    UnknownFormatError,
)
from typed_settings.loaders import (
    EnvLoader,
    FileFormat,
    FileLoader,
    InstanceLoader,
    PythonFormat,
    TomlFormat,
    clean_settings,
    tomllib,
)
from typed_settings.types import OptionList


@settings(frozen=True)
class Host:
    name: str
    port: int


@settings(frozen=True)
class Settings:
    host: Host
    url: str
    default: int = 3


class TestCleanSettings:
    """Tests for clean_settings."""

    def test_load_convert_dashes(self):
        """
        Dashes in settings and section names are replaced with underscores.
        """

        @settings(frozen=True)
        class Sub:
            b_1: str

        @settings(frozen=True)
        class Settings:
            a_1: str
            a_2: str
            sub_section: Sub

        s = {
            "a-1": "spam",
            "a_2": "eggs",
            "sub-section": {"b-1": "bacon"},
        }

        result = clean_settings(s, _deep_options(Settings), "test")
        assert result == {
            "a_1": "spam",
            "a_2": "eggs",
            "sub_section": {"b_1": "bacon"},
        }

    def test_no_replace_dash_in_dict_keys(self):
        """
        "-" in TOML keys are replaced with "_" for sections and options, but
        "-" in actuall dict keys are left alone.

        See: https://gitlab.com/sscherfke/typed-settings/-/issues/3
        """

        @settings(frozen=True)
        class Settings:
            option_1: Dict[str, Any]
            option_2: Dict[str, Any]

        s = {
            "option-1": {"my-key": "val1"},
            "option-2": {"another-key": 23},
        }

        result = clean_settings(s, _deep_options(Settings), "test")
        assert result == {
            "option_1": {"my-key": "val1"},
            "option_2": {"another-key": 23},
        }

    def test_invalid_settings(self):
        """
        Settings for which there is no attribute are errors
        """

        s = {
            "url": "abc",
            "host": {"port": 23, "eggs": 42},
            "spam": 23,
        }
        with pytest.raises(InvalidOptionsError) as exc_info:
            clean_settings(s, _deep_options(Settings), "t")
        assert str(exc_info.value) == (
            "Invalid options found in t: host.eggs, spam"
        )

    def test_clean_settings_unresolved_type(self):
        """
        Cleaning must also work if an options type is an unresolved string.
        """

        @settings(frozen=True)
        class Host:
            port: int

        @settings(frozen=True)
        class Settings:
            host: "Host"

        s = {"host": {"port": 23, "eggs": 42}}
        with pytest.raises(InvalidOptionsError) as exc_info:
            clean_settings(s, _deep_options(Settings), "t")
        assert str(exc_info.value) == "Invalid options found in t: host.eggs"

    def test_clean_settings_dict_values(self):
        """
        Some dicts may be actual values (not nested) classes.  Don't try to
        check theses as option paths.
        """

        @settings(frozen=True)
        class Settings:
            option: Dict[str, Any]

        s = {"option": {"a": 1, "b": 2}}
        clean_settings(s, _deep_options(Settings), "t")


class TestPythonFormat:
    """Tests for PythonFormat"""

    @pytest.mark.parametrize(
        "fmt, data",
        [
            (
                PythonFormat("example"),
                """\
                class example:
                    url = "spam"

                    class host:
                        port = 42
                """,
            ),
            (
                PythonFormat("EXAMPLE", key_transformer=PythonFormat.to_lower),
                """\
                class EXAMPLE:
                    URL = "spam"

                    class HOST:
                        PORT = 42
                """,
            ),
            (
                PythonFormat("example", flat=True),
                """\
                class example:
                    url = "spam"
                    host_port = 42
                """,
            ),
            (
                PythonFormat(
                    "EXAMPLE", key_transformer=PythonFormat.to_lower, flat=True
                ),
                """\
                class EXAMPLE:
                    URL = "spam"
                    HOST_PORT = 42
                """,
            ),
        ],
    )
    def test_load_python(self, fmt: FileFormat, data: str, tmp_path: Path):
        """
        We can load settings from a Python file.
        """
        config_file = tmp_path.joinpath("settings.py")
        config_file.write_text(textwrap.dedent(data))
        result = fmt(config_file, Settings, _deep_options(Settings))
        assert result == {
            "url": "spam",
            "host": {"port": 42},
        }

    @pytest.mark.parametrize("section", ["example", "spam.example"])
    def test_section_not_found(self, section: str, tmp_path: Path):
        """
        An empty dict is returned when the config file does not contain the
        desired class.
        """
        config_file = tmp_path.joinpath("settings.py")
        config_file.write_text("class spam:\n    a = 'spam'\n")
        result = PythonFormat(section)(
            config_file, Settings, _deep_options(Settings)
        )
        assert result == {}

    def test_file_not_found(self):
        """
        "ConfigFileNotFoundError" is raised when a file does not exist.
        """
        pytest.raises(
            ConfigFileNotFoundError,
            PythonFormat(""),
            Path("x"),
            _deep_options(Settings),
            Settings,
        )

    def test_file_invalid(self, tmp_path: Path):
        """
        "ConfigFileLoadError" is raised when a file contains invalid Python.
        """
        config_file = tmp_path.joinpath("settings.py")
        config_file.write_text("3x = 'spam")
        pytest.raises(
            ConfigFileLoadError,
            PythonFormat(""),
            config_file,
            _deep_options(Settings),
            Settings,
        )


class TestTomlFormat:
    """Tests for TomlFormat"""

    @pytest.mark.parametrize(
        "fmt, data",
        [
            (
                TomlFormat("example"),
                """\
                [example]
                url = "spam"
                [example.host]
                port = 42
                """,
            ),
        ],
    )
    def test_load_toml(self, fmt: FileFormat, data: str, tmp_path: Path):
        """
        We can load settings from a TOML file.
        """
        config_file = tmp_path.joinpath("settings.toml")
        config_file.write_text(textwrap.dedent(data))
        result = fmt(config_file, Settings, _deep_options(Settings))
        assert result == {
            "url": "spam",
            "host": {"port": 42},
        }

    def test_load_from_nested(self, tmp_path: Path):
        """
        We can load settings from a nested section (e.g., "tool.example").
        """
        config_file = tmp_path.joinpath("settings.toml")
        config_file.write_text(
            """[tool.example]
            a = "spam"
            [tool.example.sub]
            b = "eggs"
        """
        )
        result = TomlFormat("tool.example")(
            config_file,
            Settings,
            _deep_options(Settings),
        )
        assert result == {
            "a": "spam",
            "sub": {"b": "eggs"},
        }

    @pytest.mark.parametrize("section", ["example", "tool.example"])
    def test_section_not_found(self, section: str, tmp_path: Path):
        """
        An empty dict is returned when the config file does not contain the
        desired section.
        """
        config_file = tmp_path.joinpath("settings.toml")
        config_file.write_text(
            """[tool]
            a = "spam"
        """
        )
        result = TomlFormat(section)(
            config_file, Settings, _deep_options(Settings)
        )
        assert result == {}

    def test_file_not_found(self):
        """
        "ConfigFileNotFoundError" is raised when a file does not exist.
        """
        pytest.raises(
            ConfigFileNotFoundError,
            TomlFormat(""),
            Path("x"),
            _deep_options(Settings),
            Settings,
        )

    def test_file_not_allowed(
        self, tmp_path: Path, monkeypatch: pytest.MonkeyPatch
    ):
        """
        "ConfigFileLoadError" is raised when a file cannot be accessed.
        """

        def toml_load(path: Path):
            raise PermissionError()

        monkeypatch.setattr(tomllib, "load", toml_load)

        config_file = tmp_path.joinpath("settings.toml")
        config_file.write_text(
            """[tool]
            a = "spam"
        """
        )

        pytest.raises(
            ConfigFileLoadError,
            TomlFormat(""),
            config_file,
            _deep_options(Settings),
            Settings,
        )

    def test_file_invalid(self, tmp_path: Path):
        """
        "ConfigFileLoadError" is raised when a file contains invalid TOML.
        """
        config_file = tmp_path.joinpath("settings.toml")
        config_file.write_text("spam")
        pytest.raises(
            ConfigFileLoadError,
            TomlFormat(""),
            config_file,
            _deep_options(Settings),
            Settings,
        )


class TestFileLoader:
    """Tests for FileLoader"""

    @pytest.fixture
    def fnames(self, tmp_path: Path) -> List[Path]:
        p0 = tmp_path.joinpath("0.toml")
        p1 = tmp_path.joinpath("1.toml")
        p2 = tmp_path.joinpath("2")
        p3 = tmp_path.joinpath("3")
        p0.touch()
        p2.touch()
        return [p0, p1, p2, p3]

    @pytest.mark.parametrize(
        "cfn, env, expected",
        [
            ([], None, []),
            ([0], None, [0]),
            ([1], None, []),
            ([2], None, [2]),
            ([3], None, []),
            ([], [0], [0]),
            ([0, 1], [2, 3], [0, 2]),
            ([2, 1, 0], [2], [2, 0, 2]),
        ],
    )
    def test_get_config_filenames(
        self,
        cfn: List[int],
        env: Optional[List[int]],
        expected: List[int],
        fnames: List[Path],
        monkeypatch: MonkeyPatch,
    ):
        """
        Config files names (cfn) can be specified explicitly or via an env var.
        It's no problem if a files does not exist.
        """
        var: Optional[str]
        if env is not None:
            monkeypatch.setenv("CF", ":".join(str(fnames[i]) for i in env))
            var = "CF"
        else:
            var = None

        paths = FileLoader._get_config_filenames([fnames[i] for i in cfn], var)
        assert paths == [fnames[i] for i in expected]

    def test_get_config_filenames_empty_fn(
        self,
        fnames: List[Path],
        monkeypatch: MonkeyPatch,
    ):
        """
        Empty filenames from the env var are ignored.
        """
        monkeypatch.setenv("CF", f"::{fnames[0]}:")
        paths = FileLoader._get_config_filenames([], "CF")
        assert paths == fnames[:1]

    def test_load_file(self, tmp_path: Path):
        """
        Settings are cleaned for each file individually.
        """
        config_file = tmp_path.joinpath("settings.toml")
        config_file.write_text(
            """[le-section]
            le-option = "spam"
        """
        )

        @settings(frozen=True)
        class Settings:
            le_option: str = ""

        loader = FileLoader(
            formats={"*.toml": TomlFormat("le-section")},
            files=[config_file],
        )
        s = loader._load_file(config_file, Settings, _deep_options(Settings))
        assert s == {"le_option": "spam"}

    def test_load_file2(self, tmp_path: Path):
        """
        Settings are cleaned for each file individually.  In that process,
        "-" is normalized to "_".  This may result in duplicate settings and
        the last one wins in that case.
        """
        config_file = tmp_path.joinpath("settings.toml")
        config_file.write_text(
            """[le-section]
            le_option = "eggs"
            le-option = "spam"
        """
        )

        @settings(frozen=True)
        class Settings:
            le_option: str = ""

        loader = FileLoader(
            formats={"*.toml": TomlFormat("le-section")},
            files=[config_file],
        )
        s = loader._load_file(config_file, Settings, _deep_options(Settings))
        assert s == {"le_option": "spam"}

    def test_load_file_invalid_format(self):
        """
        An error is raised if a file has an unknown extension.
        """
        loader = FileLoader({"*.toml": TomlFormat("t")}, [])
        pytest.raises(
            UnknownFormatError, loader._load_file, Path("f.py"), [], type
        )

    def test_load(self, tmp_path: Path):
        """
        FileLoader() loads multiple files, each one overriding options
        from its predecessor.
        """
        cf1 = tmp_path.joinpath("s1.toml")
        cf1.write_text(
            """[le-section]
            le-spam = "spam"
            le-eggs = "spam"
        """
        )
        cf2 = tmp_path.joinpath("s2.toml")
        cf2.write_text(
            """[le-section]
            le_eggs = "eggs"
        """
        )

        @settings(frozen=True)
        class Settings:
            le_spam: str = ""
            le_eggs: str = ""

        loader = FileLoader({"*.toml": TomlFormat("le-section")}, [cf1, cf2])
        s = loader(Settings, _deep_options(Settings))
        assert s == {"le_spam": "spam", "le_eggs": "eggs"}

    @pytest.mark.parametrize(
        "is_mandatory, is_path, in_env, exists",
        product([True, False], repeat=4),
    )
    def test_mandatory_files(
        self,
        is_mandatory,
        is_path,
        in_env,
        exists,
        tmp_path,
        settings_cls: type,
        monkeypatch,
    ):
        """
        Paths with a "!" are mandatory and an error is raised if they don't
        exist.
        """
        p = tmp_path.joinpath("s.toml")
        if exists:
            p.touch()
        p = f"!{p}" if is_mandatory else str(p)
        if is_path:
            p = Path(p)
        files = []
        if in_env:
            monkeypatch.setenv("TEST_SETTINGS", str(p))
        else:
            files = [p]

        loader = FileLoader({"*": TomlFormat("test")}, files, "TEST_SETTINGS")
        if is_mandatory and not exists:
            pytest.raises(FileNotFoundError, loader, settings_cls, [])
        else:
            loader(settings_cls, [])


class TestEnvLoader:
    """Tests for EnvLoader"""

    def test_from_env(
        self, settings_cls: type, options: OptionList, monkeypatch: MonkeyPatch
    ):
        """
        Load options from env vars, ignore env vars for which no settings
        exist.
        """
        monkeypatch.setenv("T_URL", "foo")
        monkeypatch.setenv("T_HOST", "spam")  # Haha! Just a deceit!
        monkeypatch.setenv("T_HOST_PORT", "25")
        loader = EnvLoader(prefix="T_")
        results = loader(settings_cls, options)
        assert results == {
            "url": "foo",
            "host": {
                "port": "25",
            },
        }

    def test_no_env_prefix(
        self, settings_cls: type, options: OptionList, monkeypatch: MonkeyPatch
    ):
        """
        It is okay to use an empty prefix.
        """
        monkeypatch.setenv("URL", "spam")

        loader = EnvLoader(prefix="")
        results = loader(settings_cls, options)
        assert results == {"url": "spam"}


class TestInstanceLoader:
    """Tests for InstanceLoader"""

    def test_from_inst(
        self, settings_cls: type, options: OptionList, monkeypatch: MonkeyPatch
    ):
        """
        Load options from env vars, ignore env vars for which no settings
        exist.
        """
        # "Host" has the same attributes as the "Host" from conftest.py,
        # so it works (but only b/c it is a nested attribute and mypy doesn't
        # know what we are doing 🙈)
        inst = settings_cls(Host("spam", 42), "eggs", 23)
        loader = InstanceLoader(inst)
        results = loader(settings_cls, options)
        assert results == {
            "default": 23,
            "url": "eggs",
            "host": {
                "name": "spam",
                "port": 42,
            },
        }

    def test_invalid_type(
        self, settings_cls: type, options: OptionList, monkeypatch: MonkeyPatch
    ):
        """
        It is okay to use an empty prefix.
        """
        # "Settings" is not the same as "settings_cls" from conftest.py
        inst = Settings(Host("spam", 42), "eggs", 23)
        loader = InstanceLoader(inst)
        pytest.raises(ValueError, loader, settings_cls, options)



"""
Addtional attrs hooks
"""
from datetime import datetime
from enum import Enum
from functools import partial
from typing import TYPE_CHECKING, Any, List

import attrs
import cattrs

from ..converters import default_converter


if TYPE_CHECKING:
    try:
        from attr import Attribute, _FieldTransformer  # type: ignore
    except ImportError:
        # Just in case the symbols are moved from "attr" to "attrs"
        from attrs import Attribute, _FieldTransformer  # type: ignore


__all__ = [
    "auto_convert",
    "auto_serialize",
    "make_auto_converter",
]


def make_auto_converter(converter: cattrs.Converter) -> "_FieldTransformer":
    """
    Creates and returns an auto-converter `field transformer`_.

    .. _field transformer: https://www.attrs.org/en/stable/extending.html
                        #automatic-field-transformation-and-modification

    Args:
        converters: A cattrs :class:`~cattrs.Converter` that can handle the
            types of all fields.

    Returns:
        A function that can be passed as *field_transformer* to
        :func:`attrs.define()`.

    Example:

        .. code-block:: python

            >>> from datetime import datetime
            >>> from pathlib import Path
            >>>
            >>> import attrs
            >>> import cattrs
            >>>
            >>> converter = cattrs.GenConverter()
            >>> converter.register_structure_hook(
            ...     datetime, lambda v, _t: datetime.fromisoformat(v)
            ... )
            >>> converter.register_structure_hook(Path, lambda v, t: t(v))
            >>>
            >>> auto_convert = make_auto_converter(converter)
            >>>
            >>> @attrs.define(field_transformer=auto_convert)
            ... class C:
            ...     a: Path
            ...     b: datetime
            ...
            >>> inst = C(a="spam.md", b="2020-05-04")
            >>> inst
            C(a=PosixPath('spam.md'), b=datetime.datetime(2020, 5, 4, 0, 0))
            >>> inst.b = "2022-01-01"
            >>> inst
            C(a=PosixPath('spam.md'), b=datetime.datetime(2022, 1, 1, 0, 0))

    """

    def auto_convert(
        cls: type, attribs: List["Attribute[Any]"]
    ) -> List["Attribute[Any]"]:
        """
        A field transformer that tries to convert all attribs of a class to
        their annotated type.
        """
        attrs.resolve_types(cls, attribs=attribs)
        results = []
        for attrib in attribs:
            # Do not override explicitly defined converters!
            if attrib.converter is None:
                c = partial(converter.structure, cl=attrib.type)
                attrib = attrib.evolve(converter=c)
            results.append(attrib)

        return results

    return auto_convert


auto_convert = make_auto_converter(default_converter())
"""
An Attrs `field transformer`_ that adds converters to attributes based on their
type.

It uses the :func:`.default_converter()`.

*Deprecated:* Use ``cattrs.structure()`` instead.
"""


def auto_serialize(_inst: Any, _attrib: "Attribute[Any]", value: Any) -> Any:
    """
    Inverse hook to :func:`auto_convert` for use with :func:`attrs.asdict()`.

    *Deprecated:* Use ``cattrs.unstructure()`` instead.
    """
    if isinstance(value, datetime):
        return datetime.isoformat(value)
    if isinstance(value, Enum):
        return value.name
    return value



# pypirc_0.py

import sys
from typing import List

import typed_settings as ts


@ts.settings
class RepoServer:
    repository: str
    username: str
    password: str = ts.secret(default="")


@ts.settings
class Settings:
    index_servers: List[str]


settings = ts.load(Settings, "distutils", ["pypirc.toml"])
repos = {
    name: ts.load(RepoServer, name, ["pypirc.toml"])
    for name in settings.index_servers
}
REPO_NAME = sys.argv[1]
print(repos[REPO_NAME])



# python_gitlab.py

import typed_settings as ts


@ts.settings
class GlobalSettings:
    default: str
    ssl_verify: bool = True


@ts.settings
class GitlabAccountSettings:
    url: str
    private_token: str = ts.secret()
    api_version: int = 3


appname = "python-gitlab"
# config_files = [platformdirs.user_config_path().joinpath(f"{appname}.toml")]
config_files = [f"{appname}.toml"]

global_settings = ts.load(
    GlobalSettings,
    appname=appname,
    config_files=config_files,
    config_file_section="global",
)
account_settings = ts.load(
    GitlabAccountSettings,
    appname=appname,
    config_files=config_files,
    config_file_section=global_settings.default,
)
print(global_settings)
print(account_settings)



# pypirc_1.py

import sys
from typing import Dict

import typed_settings as ts


@ts.settings
class RepoServer:
    repository: str
    username: str
    password: str = ts.secret(default="")


@ts.settings
class Settings:
    repos: Dict[str, RepoServer]


settings = ts.load(Settings, "distutils", ["pypirc.toml"])
REPO_NAME = sys.argv[1]
print(settings.repos[REPO_NAME])
